module Set (
  Set, empty, insert, mem, union, diff
  ) where
import Data.List (intercalate)
import BTree (BTree(..)) -- want to use type without prefix
import qualified BTree

newtype Set a = Set { rep :: BTree a }

instance (Ord a, Show a) => Show (Set a) where
  show s = "{" ++ intercalate "," (elts s) ++ "}" 
    where elts = map show . BTree.flatten . rep

empty :: Set a
empty = Set BTree.Empty

insert :: Ord a => a -> Set a -> Set a
insert x s = Set (insertTree x (rep s))

mem :: Ord a => a -> Set a -> Bool
mem x s = x `memTree` (rep s)
  where memTree x Empty = False
        memTree x (Node y l r) =
          case compare x y of
            EQ -> True
            LT -> x `memTree` l
            GT -> x `memTree` r

union :: Ord a => Set a -> Set a -> Set a
union s t = Set (rep s `unionTree` rep t)

insertTree :: Ord a => a -> BTree a -> BTree a
insertTree x Empty        = Node x Empty Empty
insertTree x (Node y l r) =
  case compare x y of
    EQ -> Node y l r
    LT -> Node y (insertTree x l) r
    GT -> Node y l (insertTree x r)

unionTree :: Ord a => BTree a -> BTree a -> BTree a
unionTree Empty s        = s
unionTree (Node x l r) s =
  insertTree x (l `unionTree` r `unionTree` s)

diff :: Ord a => Set a -> Set a -> Set a
diff s t = Set (rep s `diffTree` rep t)

splitMaxTree :: BTree a -> Maybe (a,BTree a)
splitMaxTree Empty            = Nothing
splitMaxTree (Node x l Empty) = Just (x,l)
splitMaxTree (Node x l r)     =
  let Just (m,r') = splitMaxTree r
  in Just (m,Node x l r')

removeTree :: Ord a => a -> BTree a -> BTree a
removeTree x Empty        = Empty
removeTree x (Node y l r) =
  case compare x y of
    LT -> Node y (removeTree x l) r
    GT -> Node y l (removeTree x r)
    EQ -> case splitMaxTree l of
      Nothing     -> r
      Just (m,l') -> Node m l' r

diffTree :: Ord a => BTree a -> BTree a -> BTree a
diffTree t Empty        = t
diffTree t (Node x l r) = removeTree x t `diffTree` l `diffTree` r

{- Sets as lists
import qualified Data.List as List

data Set a = Set [a]

instance (Ord a, Show a) => Show (Set a) where
  show (Set xs) = "{" ++ List.intercalate "," es ++ "}"
    where es = map show (List.sort xs)

empty :: Set a
empty = Set []

insert :: Eq a => a -> Set a -> Set a
insert x (Set xs) = Set (List.nub (x : xs))

mem :: Eq a => a -> Set a -> Bool
mem x (Set xs) = x `elem` xs

union :: Eq a => Set a -> Set a -> Set a
union (Set xs) (Set ys) = Set (List.nub (xs ++ ys))

diff :: Eq a => Set a -> Set a -> Set a
diff (Set xs) (Set ys) = Set (xs List.\\ ys)
-}
